package es.ucm.fdi.grasia.faerie.broker.internal.osgi;

import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.authorityURI;
import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.localURI;
import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.nameURI;
import static es.ucm.fdi.grasia.faerie.core.util.ContextURIs.namespaceURI;

import java.io.IOException;
import java.net.URI;
import java.net.URL;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;

import org.apache.felix.scr.annotations.Activate;
import org.apache.felix.scr.annotations.Component;
import org.apache.felix.scr.annotations.Deactivate;
import org.apache.felix.scr.annotations.Reference;
import org.apache.felix.scr.annotations.ReferenceCardinality;
import org.apache.felix.scr.annotations.ReferencePolicy;
import org.apache.felix.scr.annotations.References;
import org.apache.felix.scr.annotations.Service;
import org.apache.log4j.Logger;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.service.component.ComponentContext;
import org.osgi.service.component.ComponentFactory;
import org.osgi.service.component.ComponentInstance;

import es.ucm.fdi.grasia.faerie.container.ContextContainer;
import es.ucm.fdi.grasia.faerie.core.Context;
import es.ucm.fdi.grasia.faerie.core.ContextBroker;
import es.ucm.fdi.grasia.faerie.core.Entity;
import es.ucm.fdi.grasia.faerie.core.model.BasicContextValue;
import es.ucm.fdi.grasia.faerie.core.model.BundleEntity;
import es.ucm.fdi.grasia.faerie.core.model.BundleObject;

@Service(serviceFactory = true)
@Component
@References({
		@Reference(name = "ContextContainer", referenceInterface = ContextContainer.class),
		@Reference(name = "ComponentFactory", referenceInterface = ComponentFactory.class,
				cardinality = ReferenceCardinality.OPTIONAL_MULTIPLE, policy = ReferencePolicy.DYNAMIC) })
public class OsgiContextBroker implements ContextBroker {

	private final ContextProxy contextProxy = setUpContextProxy();
	private final HashMap<String, ComponentFactory> componentFactories = setUpComponentFactories();
	private final HashMap<String, ComponentInstance> componentInstances = setUpComponentInstances();
	private ComponentContext componentContext = null;
	private BundleEntity bundleEntity = null;

	ContextProxy setUpContextProxy() {
		return new ContextProxy();
	}

	HashMap<String, ComponentFactory> setUpComponentFactories() {
		return new HashMap<String, ComponentFactory>();
	}

	HashMap<String, ComponentInstance> setUpComponentInstances() {
		return new HashMap<String, ComponentInstance>();
	}

	protected String strHashCode() {
		return Integer.toHexString(this.hashCode());
	}

	public synchronized void bindContextContainer(ContextContainer contextContainer) {
		Logger.getLogger(this.getClass()).trace("(" + contextContainer + ") * " + strHashCode());
		contextProxy.setContextContainer(contextContainer);
	}

	private String propertiesToString(Map<String, Object> properties) {
		StringBuilder builder = new StringBuilder();
		for (Map.Entry<String, Object> entry : properties.entrySet()) {
			builder.append("\n\t").append(entry.getKey()).append(" = ");
			Object value = entry.getValue();
			if (value.getClass().isArray())
				builder.append(Arrays.toString((Object[]) value));
			else
				builder.append(value);
		}
		return builder.toString();
	}

	public synchronized void bindComponentFactory(ComponentFactory factory, Map<String, Object> properties) {
		Logger.getLogger(this.getClass()).trace(
				"(" + factory + ", " + propertiesToString(properties) + ") * " + strHashCode());
		componentFactories.put((String) properties.get("component.name"), factory);
	}

	@Activate
	public synchronized void activate(ComponentContext componentContext) {
		Logger.getLogger(this.getClass()).debug(
				"(" + ContextBroker.class.getSimpleName() + ") Activating * " + strHashCode());
		this.componentContext = componentContext;
	}

	@Deactivate
	public synchronized void deactivate(ComponentContext componentContext) {
		Logger.getLogger(this.getClass()).debug(
				"(" + ContextBroker.class.getSimpleName() + ") Deactivating * " + strHashCode());
		if (bundleEntity != null) {
			contextProxy.release(bundleEntity.getId());
			bundleEntity = null;
		}
		this.componentContext = null;
	}

	public synchronized void unbindComponentFactory(ComponentFactory factory) {
		Logger.getLogger(this.getClass()).trace("(" + factory + ") * " + strHashCode());
		String[] words = factory.toString().split(" ");
		componentFactories.remove(words[2]);
	}

	public synchronized void unbindContextContainer(ContextContainer contextContainer) {
		Logger.getLogger(this.getClass()).trace("(" + contextContainer + ") * " + strHashCode());
		contextProxy.setContextContainer(null);
	}

	@Override
	public synchronized Context getContext() {
		return contextProxy;
	}

	@Override
	public synchronized BundleEntity getBundleEntity() {
		if (bundleEntity == null)
			if (componentContext != null && contextProxy != null) {
				Bundle bundle = componentContext.getUsingBundle();
				URI symName = URI.create("/" + bundle.getSymbolicName().replace(".", "/"));
				bundleEntity = contextProxy.request(new es.ucm.fdi.grasia.faerie.core.model.BundleEntity(
						authorityURI(localURI()), namespaceURI(symName), nameURI(symName)));
			}
		return bundleEntity;
	}

	@Override
	public Entity installBundle(URL bundleURL) throws IOException {
		if (componentContext != null) {
			try {
				BundleContext bundleContext = componentContext.getBundleContext();
				Bundle bundle = bundleContext.installBundle(bundleURL.toString(), bundleURL.openStream());
				URI symName = URI.create("/" + bundle.getSymbolicName().replace(".", "/"));
				es.ucm.fdi.grasia.faerie.core.model.BundleEntity installedBundleEntity = new es.ucm.fdi.grasia.faerie.core.model.BundleEntity(
						authorityURI(localURI()), namespaceURI(symName), nameURI(symName));
				BundleObject installedBundleObjectInstance = new BundleObject(installedBundleEntity);
				installedBundleObjectInstance.update(new BasicContextValue<Object>(bundle, 1.0, getBundleEntity()));
				if (!contextProxy.contains(installedBundleObjectInstance.getId())) {
					contextProxy.request(installedBundleObjectInstance);
					bundle.start();
				}
				return installedBundleEntity;
			} catch (BundleException ex) {
				Logger.getLogger(this.getClass()).error(ex.getLocalizedMessage(), ex);
			}
		}
		return null;
	}

	@Override
	public void uninstallBundle(Entity bundleEntity) {
		try {
			BundleObject componentObject = bundleEntity.getAttributeInstance(BundleObject.class);
			if (componentObject != null) {
				Bundle bundle = (Bundle) componentObject.getValue().getData();
				contextProxy.release(componentObject.getId());
				bundle.uninstall();
			}
		} catch (BundleException ex) {
			Logger.getLogger(this.getClass()).error(ex.getLocalizedMessage(), ex);
		}
	}

	@Override
	public String instantiateComponent(String componentName) {
		ComponentFactory factory = componentFactories.get(componentName);
		if (factory != null) {
			ComponentInstance instance = factory.newInstance(null);
			int hashCode = instance.hashCode();
			String code = componentName + "#" + hashCode;
			componentInstances.put(code, instance);
			return code;
		}
		return null;
	}

	@Override
	public void disposeComponent(String componentObject) {
		ComponentInstance componentInstance = componentInstances.get(componentObject);
		if (componentInstance != null)
			componentInstance.dispose();
	}
}
